# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import functools
import graphviz
import os
import re
import subprocess

from trac.config import BoolOption, ChoiceOption, ListOption, Option
from trac.core import Component, TracError, implements
from trac.resource import ResourceNotFound, get_resource_summary
from trac.ticket.model import Ticket
from trac.ticket.query import Query
from trac.util import as_int, escape, to_unicode
from trac.util.html import html
from trac.util.presentation import classes
from trac.util.translation import _, tag_
from trac.web.api import IRequestFilter, IRequestHandler, \
                         ITemplateStreamFilter
from trac.web.chrome import ITemplateProvider, add_ctxtnav, add_script

from model import TicketLinks


class MasterTicketsModule(Component):
    """Provides support for ticket dependencies."""

    implements(IRequestFilter, IRequestHandler, ITemplateProvider,
               ITemplateStreamFilter)

    dot_path = Option('mastertickets', 'dot_path', default='dot',
        doc="Path to the dot executable.")

    gs_path = Option('mastertickets', 'gs_path', default='gs',
        doc="Path to the ghostscript executable.")

    use_gs = BoolOption('mastertickets', 'use_gs', default=False,
        doc="If enabled, use ghostscript to produce nicer output.")

    acceptable_formats = ListOption('mastertickets', 'acceptable_formats',
                                    default='png,cmapx', sep=',',
        doc="""The formats that may be chosen; execute dot -T? for a
            list of options.""")

    closed_color = Option('mastertickets', 'closed_color', default='green',
        doc="Color of closed tickets")

    opened_color = Option('mastertickets', 'opened_color', default='red',
        doc="Color of opened tickets")

    show_key = Option('mastertickets', 'show_key', default=False,
        doc="Show a key for open/closed nodes")

    closed_text = Option('mastertickets', 'closed_text', default='Done',
        doc="Text for key showing closed tickets")

    opened_text = Option('mastertickets', 'opened_text', default='ToDo',
        doc="Text for key showing opened tickets")

    highlight_target = Option('mastertickets', 'highlight_target',
                              default=False,
        doc="Highlight target tickets in graph")

    full_graph = Option('mastertickets', 'full_graph', default=False,
        doc="Show full dep. graph, not just direct blocking links")

    graph_direction = ChoiceOption('mastertickets', 'graph_direction',
                                   choices=['TD', 'LR', 'DT', 'RL'],
        doc="""Direction of the dependency graph (TD = Top Down,
            DT = Down Top, LR = Left Right, RL = Right Left).""")

    fields = set(['blocking', 'blockedby'])

    # IRequestFilter methods

    def pre_process_request(self, req, handler):
        return handler

    def post_process_request(self, req, template, data, content_type):
        if template is not None:
            if req.path_info.startswith('/ticket/'):
                # In case of an invalid ticket, the data is invalid
                if not data:
                    return template, data, content_type
                tkt = data['ticket']
                links = TicketLinks(self.env, tkt)

                for i in links.blocked_by:
                    if Ticket(self.env, i)['status'] != 'closed':
                        add_script(req, 'mastertickets/js/disable_resolve.js')
                        break

                # Add link to depgraph if needed.
                if links:
                    add_ctxtnav(req, 'Depgraph',
                                req.href.depgraph('ticket', tkt.id))

                for change in data.get('changes', {}):
                    if 'fields' not in change:
                        continue
                    for field, field_data in change['fields'].iteritems():
                        if field in self.fields:
                            if field_data['new'].strip():
                                new = to_int_set(field_data['new'])
                            else:
                                new = set()
                            if field_data['old'].strip():
                                old = to_int_set(field_data['old'])
                            else:
                                old = set()
                            add = new - old
                            sub = old - new
                            elms = html()
                            if add:
                                elms.append(
                                    html.em(u', '.join(unicode(n)
                                                       for n in sorted(add)))
                                )
                                elms.append(u' added')
                            if add and sub:
                                elms.append(u'; ')
                            if sub:
                                elms.append(
                                    html.em(u', '.join(unicode(n)
                                                       for n in sorted(sub)))
                                )
                                elms.append(u' removed')
                            field_data['rendered'] = elms

            # Add a link to generate a dependency graph for all the tickets
            # in the milestone
            if req.path_info.startswith('/milestone/'):
                if not data:
                    return template, data, content_type
                milestone = data['milestone']
                add_ctxtnav(req, 'Depgraph',
                            req.href.depgraph('milestone', milestone.name))

        return template, data, content_type

    # ITemplateStreamFilter methods

    def filter_stream(self, req, method, filename, stream, data):
        if not data:
            return stream

        # Try all at the same time to catch changed or processed templates.
        if filename in ['report_view.html', 'query_results.html',
                        'ticket.html', 'query.html']:
            # For ticket.html
            if 'fields' in data and isinstance(data['fields'], list):
                for field in data['fields']:
                    for f in self.fields:
                        if field['name'] == f and data['ticket'][f]:
                            field['rendered'] = \
                                self._link_tickets(req, data['ticket'][f])
            # For query_results.html and query.html
            if 'groups' in data and isinstance(data['groups'], list):
                for group, tickets in data['groups']:
                    for ticket in tickets:
                        for f in self.fields:
                            if f in ticket:
                                ticket[f] = self._link_tickets(req, ticket[f])
            # For report_view.html
            if 'row_groups' in data and isinstance(data['row_groups'], list):
                for group, rows in data['row_groups']:
                    for row in rows:
                        if 'cell_groups' in row and \
                                isinstance(row['cell_groups'], list):
                            for cells in row['cell_groups']:
                                for cell in cells:
                                    # If the user names column in the report
                                    # differently (blockedby AS "blocked by")
                                    # then this will not find it
                                    if cell.get('header', {}).get('col') \
                                            in self.fields:
                                        cell['value'] = \
                                            self._link_tickets(req,
                                                               cell['value'])
        return stream

    # ITemplateProvider methods

    def get_htdocs_dirs(self):
        from pkg_resources import resource_filename
        return [('mastertickets', resource_filename(__name__, 'htdocs'))]

    def get_templates_dirs(self):
        from pkg_resources import resource_filename
        return [resource_filename(__name__, 'templates')]

    # IRequestHandler methods

    def match_request(self, req):
        match = re.match(r'^/depgraph/(?P<realm>ticket|milestone)/'
                         r'(?P<id>((?!depgraph.png).)+)(/depgraph.png)?$',
                         req.path_info)
        if match:
            req.args['realm'] = match.group('realm')
            req.args['id'] = match.group('id')
            return True

    def process_request(self, req):
        realm = req.args['realm']
        id_ = req.args['id']

        if not which(self.dot_path):
            raise TracError(_("Path to dot executable is invalid: %(path)s",
                              path=self.dot_path))

        # Urls to generate the depgraph for a ticket is /depgraph/ticketnum
        # Urls to generate the depgraph for a milestone is
        # /depgraph/milestone/milestone_name

        # List of tickets to generate the depgraph.
        if realm == 'milestone':
            # We need to query the list of tickets in the milestone
            query = Query(self.env, constraints={'milestone': [id_]}, max=0)
            tkt_ids = [fields['id'] for fields in query.execute(req)]
        else:
            tid = as_int(id_, None)
            if tid is None:
                raise TracError(tag_("%(id)s is not a valid ticket id.",
                                     id=html.tt(id_)))
            tkt_ids = [tid]

        # The summary argument defines whether we place the ticket id or
        # its summary in the node's label
        label_summary = 0
        if 'summary' in req.args:
            label_summary = int(req.args.get('summary'))

        g = self._build_graph(req, tkt_ids, label_summary=label_summary)
        if req.path_info.endswith('/depgraph.png') or 'format' in req.args:
            format_ = req.args.get('format')
            if format_ == 'text':
                # In case g.__str__ returns unicode, convert it in ascii
                req.send(to_unicode(g).encode('ascii', 'replace'),
                         'text/plain')
            elif format_ == 'debug':
                import pprint

                req.send(
                    pprint.pformat(
                        [TicketLinks(self.env, tkt_id) for tkt_id in tkt_ids]
                    ),
                    'text/plain')
            elif format_ is not None:
                if format_ in self.acceptable_formats:
                    req.send(g.render(self.dot_path, format_), 'text/plain')
                else:
                    raise TracError(_("The %(format)s format is not allowed.",
                                      format=format_))

            if self.use_gs:
                ps = g.render(self.dot_path, 'ps2')
                gs = subprocess.Popen(
                    [self.gs_path, '-q', '-dTextAlphaBits=4',
                     '-dGraphicsAlphaBits=4', '-sDEVICE=png16m',
                     '-sOutputFile=%stdout%', '-'],
                    stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
                img, err = gs.communicate(ps)
                if err:
                    self.log.debug('MasterTickets: Error from gs: %s', err)
            else:
                img = g.render(self.dot_path)
            req.send(img, 'image/png')
        else:
            data = {}

            # Add a context link to enable/disable labels in nodes.
            if label_summary:
                add_ctxtnav(req, 'Without labels',
                            req.href(req.path_info, summary=0))
            else:
                add_ctxtnav(req, 'With labels',
                            req.href(req.path_info, summary=1))

            if realm == 'milestone':
                add_ctxtnav(req, 'Back to Milestone: %s' % id_,
                            req.href.milestone(id_))
                data['milestone'] = id_
            else:
                data['ticket'] = id_
                add_ctxtnav(req, 'Back to Ticket #%s' % id_,
                            req.href.ticket(id_))
            data['graph'] = g
            data['graph_render'] = functools.partial(g.render, self.dot_path)
            data['use_gs'] = self.use_gs

            return 'depgraph.html', data, None

    def _build_graph(self, req, tkt_ids, label_summary=0):
        g = graphviz.Graph(log=self.log)
        g.label_summary = label_summary

        g.attributes['rankdir'] = self.graph_direction

        node_default = g['node']
        node_default['style'] = 'filled'

        edge_default = g['edge']
        edge_default['style'] = ''

        # Force this to the top of the graph
        for tid in tkt_ids:
            g[tid]

        if self.show_key:
            g[-1]['label'] = self.closed_text
            g[-1]['fillcolor'] = self.closed_color
            g[-1]['shape'] = 'box'
            g[-2]['label'] = self.opened_text
            g[-2]['fillcolor'] = self.opened_color
            g[-2]['shape'] = 'box'

        links = TicketLinks.walk_tickets(self.env, tkt_ids, self.full_graph)
        links = sorted(links, key=lambda link: link.tkt.id)
        for link in links:
            tkt = link.tkt
            node = g[tkt.id]
            if label_summary:
                node['label'] = u'#%s %s' % (tkt.id, tkt['summary'])
            else:
                node['label'] = u'#%s' % tkt.id
            node['fillcolor'] = tkt['status'] == 'closed' and \
                                self.closed_color or self.opened_color
            node['URL'] = req.href.ticket(tkt.id)
            node['alt'] = u'Ticket #%s' % tkt.id
            node['tooltip'] = escape(tkt['summary'])
            if self.highlight_target and tkt.id in tkt_ids:
                node['penwidth'] = 3

            for n in link.blocking:
                node > g[n]

        return g

    def _link_tickets(self, req, tickets):
        items = []

        for i, word in enumerate(re.split(r'([;,\s]+)', tickets)):
            if i % 2:
                items.append(word)
            elif word:
                tid = word
                word = '#%s' % word

                try:
                    ticket = Ticket(self.env, tid)
                    if 'TICKET_VIEW' in req.perm(ticket.resource):
                        word = \
                            html.a(
                                '#%s' % ticket.id,
                                href=req.href.ticket(ticket.id),
                                class_=classes(ticket['status'], 'ticket'),
                                title=get_resource_summary(self.env,
                                                           ticket.resource)
                            )
                except ResourceNotFound:
                    pass

                items.append(word)

        if items:
            return html(items)
        else:
            return None


def to_int_set(splittable):
    return set(int(n) for n in re.split(r'[\s,]+', splittable))


def which(program):

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import re

from trac.core import Component, implements
from trac.db import DatabaseManager
from trac.env import IEnvironmentSetupParticipant
from trac.ticket.api import ITicketChangeListener, ITicketManipulator
from trac.ticket.model import Ticket

from mastertickets import compat
from mastertickets import db_default
from mastertickets.model import TicketLinks


class MasterTicketsSystem(Component):
    """Central functionality for the MasterTickets plugin."""

    implements(IEnvironmentSetupParticipant, ITicketChangeListener,
               ITicketManipulator)

    NUMBERS_RE = re.compile(r'\d+', re.U)

    # IEnvironmentSetupParticipant methods

    def environment_created(self):
        with self.env.db_transaction as db:
            self.upgrade_environment(db)

    def environment_needs_upgrade(self, db):
        dbm = DatabaseManager(self.env)
        return dbm.needs_upgrade(db_default.version, db_default.name) or \
               self._ticket_custom_needs_upgrade()

    def upgrade_environment(self, db):
        dbm = DatabaseManager(self.env)
        if dbm.needs_upgrade(db_default.version, db_default.name):
            if not dbm.get_database_version(db_default.name):
                dbm.create_tables(db_default.tables)
                dbm.set_database_version(db_default.version, db_default.name)
            else:
                dbm.upgrade(db_default.version, db_default.name,
                            'mastertickets.upgrades')

        custom = self.config['ticket-custom']
        save_config = False
        if 'blocking' not in custom:
            custom.set('blocking', 'text')
            custom.set('blocking.label', 'Blocking')
            save_config = True
        if 'blockedby' not in custom:
            custom.set('blockedby', 'text')
            custom.set('blockedby.label', 'Blocked By')
            save_config = True
        if save_config:
            self.config.save()

    # ITicketChangeListener methods

    def ticket_created(self, tkt):
        self.ticket_changed(tkt, '', tkt['reporter'], {})

    def ticket_changed(self, tkt, comment, author, old_values):
        links = self._prepare_links(tkt)
        links.save(author, comment, tkt.time_changed)

    def ticket_deleted(self, tkt):
        links = TicketLinks(self.env, tkt)
        links.blocking = set()
        links.blocked_by = set()
        links.save('trac', "Ticket #%s deleted" % tkt.id)

    # ITicketManipulator methods

    def prepare_ticket(self, req, ticket, fields, actions):
        pass

    def validate_ticket(self, req, ticket):
        tid = ticket.id
        links = self._prepare_links(ticket)

        if req.args.get('action') == 'resolve' and \
                req.args.get('action_resolve_resolve_resolution') == 'fixed':
            for i in links.blocked_by:
                if Ticket(self.env, i)['status'] != 'closed':
                    yield None, "Ticket #%s is blocking this ticket" % i

        # Check that ticket does not have itself as a blocker
        if tid in links.blocking | links.blocked_by:
            yield 'blocked_by', "This ticket is blocking itself"
            return

        # Check that there aren't any blocked_by in blocking or their parents
        blocking = links.blocking.copy()
        while len(blocking) > 0:
            if len(links.blocked_by & blocking) > 0:
                yield 'blocked_by', "This ticket has circular dependencies"
                return
            new_blocking = set()
            for link in blocking:
                tmp_tkt = Ticket(self.env, link)
                new_blocking |= TicketLinks(self.env, tmp_tkt).blocking
            blocking = new_blocking

        for field in ('blocking', 'blockedby'):
            try:
                ids = self.NUMBERS_RE.findall(ticket[field] or '')
                for tid in ids[:]:
                    for _ in self.env.db_query("""
                            SELECT id FROM ticket WHERE id=%s
                            """, (tid,)):
                        break
                    else:
                        ids.remove(tid)
                ticket[field] = ', '.join(sorted(ids, key=lambda x: int(x)))
            except Exception, e:
                self.log.debug("MasterTickets: Error parsing %s \"%s\": %s",
                               field, ticket[field], e)
                yield field, "Not a valid list of ticket IDs"

    # Internal methods

    def _prepare_links(self, tkt):
        links = TicketLinks(self.env, tkt)
        links.blocking = set(int(n) for n in self.NUMBERS_RE.findall(tkt['blocking'] or ''))
        links.blocked_by = set(int(n) for n in self.NUMBERS_RE.findall(tkt['blockedby'] or ''))
        return links

    def _ticket_custom_needs_upgrade(self):
        return 'blocking' not in self.config['ticket-custom'] or \
               'blockedby' not in self.config['ticket-custom']



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

from trac.core import TracError
from trac.db.api import DatabaseManager
from trac.db.schema import Table
from trac.util.translation import _

try:
    from trac.util import to_list  # Trac 1.0.3 and later
except ImportError:
    def to_list(splittable, sep=','):
        """Split a string at `sep` and return a list without any empty items.
        """
        split = [x.strip() for x in splittable.split(sep)]
        return [item for item in split if item]


# Methods backported from Trac 1.2

if not hasattr(DatabaseManager, 'drop_tables'):
    def drop_tables(self, schema):
        """Drop the specified tables.

        :param schema: an iterable of `Table` objects or table names.

        :since: version 1.0.2
        """
        with self.env.db_transaction as db:
            for table in schema:
                table_name = table.name if isinstance(table, Table) else table
                db.drop_table(table_name)
    DatabaseManager.drop_tables = drop_tables


if not hasattr(DatabaseManager, 'create_tables'):
    def create_tables(self, schema):
        """Create the specified tables.

        :param schema: an iterable of table objects.

        :since: version 1.0.2
        """
        connector = self.get_connector()[0]
        with self.env.db_transaction as db:
            for table in schema:
                for sql in connector.to_sql(table):
                    db(sql)
    DatabaseManager.create_tables = create_tables


if not hasattr(DatabaseManager, 'get_database_version'):
    def get_database_version(self, name='database_version'):
        """Returns the database version from the SYSTEM table as an int,
        or `False` if the entry is not found.

        :param name: The name of the entry that contains the database version
                     in the SYSTEM table. Defaults to `database_version`,
                     which contains the database version for Trac.
        """
        rows = self.env.db_query("""
                SELECT value FROM system WHERE name=%s
                """, (name,))
        return int(rows[0][0]) if rows else False
    DatabaseManager.get_database_version = get_database_version


if not hasattr(DatabaseManager, 'set_database_version'):
    def set_database_version(self, version, name='database_version'):
        """Sets the database version in the SYSTEM table.

        :param version: an integer database version.
        :param name: The name of the entry that contains the database version
                     in the SYSTEM table. Defaults to `database_version`,
                     which contains the database version for Trac.
        """
        current_database_version = self.get_database_version(name)
        if current_database_version is False:
            self.env.db_transaction("""
                    INSERT INTO system (name, value) VALUES (%s, %s)
                    """, (name, version))
        else:
            self.env.db_transaction("""
                    UPDATE system SET value=%s WHERE name=%s
                    """, (version, name))
            self.log.info("Upgraded %s from %d to %d",
                          name, current_database_version, version)
    DatabaseManager.set_database_version = set_database_version


if not hasattr(DatabaseManager, 'get_table_names'):
    def get_table_names(self):
        dburi = self.config.get('trac', 'database')
        if dburi.startswith('sqlite:'):
            query = "SELECT name FROM sqlite_master" \
                    " WHERE type='table' AND NOT name='sqlite_sequence'"
        elif dburi.startswith('postgres:'):
            query = "SELECT tablename FROM pg_tables" \
                    " WHERE schemaname = ANY (current_schemas(false))"
        elif dburi.startswith('mysql:'):
            query = "SHOW TABLES"
        else:
            raise TracError('Unsupported %s database' % dburi.split(':')[0])
        return sorted(row[0] for row in self.env.db_transaction(query))


if not hasattr(DatabaseManager, 'needs_upgrade'):
    def needs_upgrade(self, version, name='database_version'):
        """Checks the database version to determine if an upgrade is needed.

        :param version: the expected integer database version.
        :param name: the name of the entry in the SYSTEM table that contains
                     the database version. Defaults to `database_version`,
                     which contains the database version for Trac.

        :return: `True` if the stored version is less than the expected
                  version, `False` if it is equal to the expected version.
        :raises TracError: if the stored version is greater than the expected
                           version.
        """
        dbver = self.get_database_version(name)
        if dbver == version:
            return False
        elif dbver > version:
            raise TracError(_("Need to downgrade %(name)s.", name=name))
        self.log.info("Need to upgrade %s from %d to %d",
                      name, dbver, version)
        return True
    DatabaseManager.needs_upgrade = needs_upgrade


if not hasattr(DatabaseManager, 'upgrade'):
    def upgrade(self, version, name='database_version', pkg=None):
        """Invokes `do_upgrade(env, version, cursor)` in module
        `"%s/db%i.py" % (pkg, version)`, for each required version upgrade.

        :param version: the expected integer database version.
        :param name: the name of the entry in the SYSTEM table that contains
                     the database version. Defaults to `database_version`,
                     which contains the database version for Trac.
        :param pkg: the package containing the upgrade modules.

        :raises TracError: if the package or module doesn't exist.
        """
        dbver = self.get_database_version(name)
        for i in range(dbver + 1, version + 1):
            module = 'db%i' % i
            try:
                upgrades = __import__(pkg, globals(), locals(), [module])
            except ImportError:
                raise TracError(_("No upgrade package %(pkg)s", pkg=pkg))
            try:
                script = getattr(upgrades, module)
            except AttributeError:
                raise TracError(_("No upgrade module %(module)s.py",
                                  module=module))
            with self.env.db_transaction as db:
                cursor = db.cursor()
                script.do_upgrade(self.env, i, cursor)
                self.set_database_version(i, name)
    DatabaseManager.upgrade = upgrade



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import copy
import datetime

from trac.ticket.model import Ticket
from trac.util.datefmt import utc, to_utimestamp

from mastertickets.compat import to_list


class TicketLinks(object):
    """A model for the ticket links used MasterTickets."""

    def __init__(self, env, tkt):
        self.env = env
        if not isinstance(tkt, Ticket):
            tkt = Ticket(self.env, tkt)
        self.tkt = tkt

        self.blocking = set()
        for dest, in self.env.db_query("""
                SELECT dest FROM mastertickets WHERE source=%s ORDER BY dest
                """, (self.tkt.id,)):
            self.blocking.add(int(dest))
        self._old_blocking = copy.copy(self.blocking)

        self.blocked_by = set()
        for source, in self.env.db_query("""
                SELECT source FROM mastertickets WHERE dest=%s ORDER BY source
                """, (self.tkt.id,)):
            self.blocked_by.add(int(source))
        self._old_blocked_by = copy.copy(self.blocked_by)

    def save(self, author, comment='', when=None):
        """Save new links."""
        if when is None:
            when = datetime.datetime.now(utc)
        when_ts = to_utimestamp(when)

        new_blocking = set(int(n) for n in self.blocking)
        new_blocked_by = set(int(n) for n in self.blocked_by)

        to_check = [  # new, old, field
            (new_blocking, self._old_blocking, 'blockedby',
             ('source', 'dest')),
            (new_blocked_by, self._old_blocked_by, 'blocking',
             ('dest', 'source')),
        ]

        def append_id(lst):
            lst.append(str(self.tkt.id))

        def remove_id(lst):
            lst.remove(str(self.tkt.id))

        with self.env.db_transaction as db:
            for new_ids, old_ids, field, sourcedest in to_check:
                for n in new_ids | old_ids:
                    update_field = None
                    if n in new_ids and n not in old_ids:
                        # New ticket added
                        db("""INSERT INTO mastertickets (%s, %s)
                            VALUES (%%s, %%s)
                            """ % sourcedest, (self.tkt.id, n))
                        update_field = append_id
                    elif n not in new_ids and n in old_ids:
                        # Old ticket removed
                        db("""
                            DELETE FROM mastertickets WHERE %s=%%s AND %s=%%s
                            """ % sourcedest, (self.tkt.id, n))
                        update_field = remove_id

                    if update_field is not None:
                        old_value = None
                        ticket_custom_row_exists = False
                        for old_value, in db("""
                                SELECT value FROM ticket_custom
                                WHERE ticket=%s AND name=%s
                                """, (n, field)):
                            ticket_custom_row_exists = True
                            break
                        if old_value:
                            new_value = to_list(old_value)
                        else:
                            new_value = []
                        update_field(new_value)
                        new_value = ', '.join(sorted(new_value,
                                                     key=lambda x: int(x)))

                        # ticket, time and field must be unique for database
                        # integrity. The TicketImportPlugin assigns the same
                        # changetime to all ticket if not specified, which
                        # was causing an IntegrityError (#10194).
                        changelog = Ticket(self.env, n).get_changelog(when)
                        if any(field in cl for cl in changelog):
                            db("""UPDATE ticket_change
                                  SET author=%s, oldvalue=%s, newvalue=%s
                                  WHERE ticket=%s AND time=%s AND field=%s
                                  """, (author, old_value, new_value, n,
                                        when_ts, field))
                        else:
                            db("""INSERT INTO ticket_change
                                   (ticket, time, author, field,
                                    oldvalue, newvalue)
                                  VALUES (%s, %s, %s, %s, %s, %s)
                                  """, (n, when_ts, author, field,
                                        old_value, new_value))

                        # Add comment to referenced ticket if a comment
                        # hasn't already been added
                        if comment and \
                                not any('comment' in entry
                                        for entry
                                        in self.tkt.get_changelog(when)):
                            db("""INSERT INTO ticket_change
                                   (ticket, time, author, field,
                                    oldvalue, newvalue)
                                  VALUES (%s, %s, %s, %s, %s, %s)
                                  """, (n, when_ts, author, 'comment', '',
                                        '(In #%s) %s'
                                        % (self.tkt.id, comment)))

                        if ticket_custom_row_exists:
                            db("""UPDATE ticket_custom SET value=%s
                                  WHERE ticket=%s AND name=%s
                                  """, (new_value, n, field))
                        else:
                            db("""INSERT INTO ticket_custom
                                   (ticket, name, value)
                                  VALUES (%s, %s, %s)
                                  """, (n, field, new_value))

                        # Refresh the changetime to prevent concurrent edits.
                        db("UPDATE ticket SET changetime=%s WHERE id=%s",
                           (when_ts, n))

        # cursor.execute('DELETE FROM mastertickets
        #                'WHERE source=%s OR dest=%s',
        #                (self.tkt.id, self.tkt.id))
        # data = []
        # for tkt in self.blocking:
        #     if isinstance(tkt, Ticket):
        #         tkt = tkt.id
        #     data.append((self.tkt.id, tkt))
        # for tkt in self.blocked_by:
        #     if isinstance(tkt, Ticket):
        #         tkt = tkt.id
        #     data.append((tkt, self.tkt.id))
        #
        # cursor.executemany('INSERT INTO mastertickets (source, dest) '
        #                    'VALUES (%s, %s)', data)

    def __nonzero__(self):
        return bool(self.blocking) or bool(self.blocked_by)

    def __repr__(self):
        def l(arr):
            arr2 = []
            for tkt in arr:
                if isinstance(tkt, Ticket):
                    tkt = tkt.id
                arr2.append(str(tkt))
            return '[%s]' % ','.join(arr2)

        return '<mastertickets.model.TicketLinks #%s ' \
               'blocking=%s blocked_by=%s>' % \
               (self.tkt.id, l(getattr(self, 'blocking', [])),
                l(getattr(self, 'blocked_by', [])))

    @staticmethod
    def walk_tickets(env, tkt_ids, full=False):
        """Return an iterable of all links reachable directly above or
        below those ones.
        """

        def visit(tkt, memo, next_fn):
            if tkt in memo:
                return False

            links = TicketLinks(env, tkt)
            memo[tkt] = links

            for n in next_fn(links):
                visit(n, memo, next_fn)

        memo1 = {}
        memo2 = {}
        for tid in tkt_ids:
            if full:
                visit(tid, memo1, lambda links: links.blocking |
                                               links.blocked_by)
            else:
                visit(tid, memo1, lambda links: links.blocking)
                visit(tid, memo2, lambda links: links.blocked_by)
        memo1.update(memo2)
        return memo1.itervalues()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

from trac.db import Table, Column

name = 'mastertickets'
version = 2
tables = [
    Table('mastertickets', key=('source', 'dest'))[
        Column('source', type='integer'),
        Column('dest', type='integer'),
    ],
]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

from trac.db.api import DatabaseManager

from mastertickets import db_default


def do_upgrade(env, ver, cursor):
    dbm = DatabaseManager(env)
    temp_table_name = db_default.name + '_old'
    with env.db_transaction as db:
        db("""CREATE TEMPORARY TABLE %s AS SELECT * FROM %s
            """ % (temp_table_name, db_default.name))
        dbm.drop_tables((db_default.name,))
        dbm.create_tables(db_default.tables)
        for source, dest in db("""
                SELECT source, dest FROM %s
                """ % temp_table_name):
            try:
                db("""
                    INSERT INTO %s (source,dest) VALUES (%%s,%%s)
                    """ % db_default.name, (int(source), int(dest)))
            except Exception:
                env.log.warning("Error inserting source, dest = %s, %s",
                                source, dest)
        dbm.drop_tables((temp_table_name,))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import shutil
import tempfile
import unittest

from trac.test import EnvironmentStub, Mock


class TicketLinksTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(default_data=True,
                                   enable=['trac.*', 'mastertickets.*'])
        self.env.path = tempfile.mkdtemp()
        self.req = Mock()

    def tearDown(self):
        self.env.shutdown()
        shutil.rmtree(self.env.path)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TicketLinksTestCase, 'test'))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import itertools
import subprocess

from trac.util.text import to_unicode
from trac.util.translation import _


def _format_options(base_string, options):
    return u'%s [%s]' \
           % (base_string,
              u', '.join(u'%s="%s"' % x for x in options.iteritems()))


class Edge(dict):
    """Model for an edge in a dot graph."""

    def __init__(self, source, dest, **kwargs):
        self.source = source
        self.dest = dest
        dict.__init__(self, **kwargs)

    def __str__(self):
        ret = u'%s -> %s' % (self.source.name, self.dest.name)
        if self:
            ret = _format_options(ret, self)
        return ret

    def __hash__(self):
        return hash(id(self))


class Node(dict):
    """Model for a node in a dot graph."""

    def __init__(self, name, **kwargs):
        self.name = unicode(name)
        self.edges = []
        dict.__init__(self, **kwargs)

    def __str__(self):
        ret = self.name
        if self:
            ret = _format_options(ret, self)
        return ret

    def __gt__(self, other):
        """Allow node1 > node2 to add an edge."""
        edge = Edge(self, other)
        self.edges.append(edge)
        other.edges.append(edge)
        return edge

    def __lt__(self, other):
        edge = Edge(other, self)
        self.edges.append(edge)
        other.edges.append(edge)
        return edge

    def __hash__(self):
        return hash(id(self))


class Graph(object):
    """A model object for a graphviz digraph."""

    def __init__(self, name=u'graph', log=None):
        super(Graph, self).__init__()
        self.name = name
        self.log = log
        self.nodes = []
        self._node_map = {}
        self.attributes = {}
        self.edges = []

    def add(self, obj):
        if isinstance(obj, Node):
            self.nodes.append(obj)
            self._node_map[obj.name] = obj
        elif isinstance(obj, Edge):
            self.edges.append(obj)

    def __getitem__(self, key):
        key = unicode(key)
        if key not in self._node_map:
            new_node = Node(key)
            self._node_map[key] = new_node
            self.nodes.append(new_node)
        return self._node_map[key]

    def __delitem__(self, key):
        key = unicode(key)
        node = self._node_map.pop(key)
        self.nodes.remove(node)

    def __str__(self):
        edges = []
        nodes = []

        memo = set()

        def process(lst):
            for item in lst:
                if item in memo:
                    continue
                memo.add(item)

                if isinstance(item, Node):
                    nodes.append(item)
                    process(item.edges)
                elif isinstance(item, Edge):
                    edges.append(item)
                    if isinstance(item.source, Node):
                        process((item.source,))
                    if isinstance(item.dest, Node):
                        process((item.dest,))

        process(self.nodes)
        process(self.edges)

        lines = [u'digraph "%s" {' % self.name]
        for att, value in self.attributes.iteritems():
            lines.append(u'\t%s="%s";' % (att, value))
        for obj in itertools.chain(nodes, edges):
            lines.append(u'\t%s;' % obj)
        lines.append(u'}')
        return u'\n'.join(lines)

    def render(self, dot_path='dot', format='png'):
        """Render a dot graph."""
        cmd = [dot_path, '-T%s' % format]
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        out, error = p.communicate(to_unicode(self).encode('utf8'))
        if error or p.returncode and self.log:
            self.log.error(_("dot %(dot_path)s failed with code %(rc)s: "
                             "%(error)s", dot_path=dot_path, rc=p.returncode,
                             error=error))
        return out


if __name__ == '__main__':
    g = Graph()
    root = Node('me')
    root > Node('them')
    root < Node(u'Üs')

    g.add(root)

    print g.render()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import unittest


def suite():
    from mastertickets.tests import model
    suite = unittest.TestSuite()
    suite.addTest(model.suite())
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2012 Noah Kantrowitz <noah@coderanger.net>
# Copyright (c) 2013-2016 Ryan J Ollos <ryan.j.ollos@gmail.com>
#
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.
#

import os
from setuptools import find_packages, setup

setup(
    name='TracMasterTickets',
    version='4.0.1',
    packages=find_packages(exclude=['*.tests']),
    package_data={
        'mastertickets': [
            'htdocs/img/*.gif',
            'htdocs/img/*.png',
            'htdocs/js/*.js',
            'htdocs/css/*.css',
            'templates/*.html'
        ]
    },

    author='Noah Kantrowitz',
    author_email='noah@coderanger.net',
    maintainer='Ryan J Ollos',
    maintainer_email='ryan.j.ollos@gmail.com',
    description='Provides ticket dependencies and master tickets.',
    long_description=open(os.path.join(os.path.dirname(__file__),
                                       'README')).read(),
    license='BSD 3-Clause',
    keywords='trac plugin ticket dependencies master',
    url='https://trac-hacks.org/wiki/MasterTicketsPlugin',
    classifiers=[
        'Framework :: Trac',
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],

    install_requires=['Trac'],
    test_suite='mastertickets.tests.suite',
    entry_points={
        'trac.plugins': [
            'mastertickets.api = mastertickets.api',
            'mastertickets.web_ui = mastertickets.web_ui',
        ]
    }
)
